
/**
* Base methods for auto generated routes.
*
* @module actions
* @param {object} xipframe - loaded Xipframe instance.
* @returns {object}
*/
module.exports = function actions(xipframe) {

   'use strict';

    var config = xipframe.config
      , utils = xipframe.utils
      , _ = utils._
      , DatabaseError = xipframe.errors.DatabaseError
      , DatabaseValidationError = xipframe.errors.DatabaseValidationError
      , NotImplementedError = xipframe.errors.NotImplmentedError
      , NullObjectError = xipframe.errors.NullObjectError
      , nullObject = 'Null object reference missing parameter or post body.'   
      , notFound = 'Model not found check filter parameters.'
      , undefModel = 'Cannot save model of type undefined.'
      , viewsPath = config.paths.views + '/'
      , viewsExt = '.' + config.viewEngine
      , url = require('url');
    

    /*
    * @function call
    * @param {object} options - additional route options.  
    * @augments actions
    */
    return function call(options) {

        var Model = options.model
          , xModel = options.xModel
          , query = ''
          , lastOrder = 'id'
          , lastDir = 'ASC'
          , template
          , viewDir
          , viewType
          , viewValid;
        
        if (options.view) {
           
            viewDir = options.view.split('/')[0];
            viewType = options.view.split('/')[1];
            viewValid = _.contains(config.autoRoute.views, viewType);   

            /* check if view exists if not create it */
            if (config.autoRoute.createViews && viewValid && !utils.diskio.existsSync(viewsPath + options.view + viewsExt)) {

                /* check that we have a valid views directory */
                if (!utils.diskio.existsSync(viewsPath + viewDir))
                    utils.diskio.mkdirSync(viewsPath + viewDir);

                /* validate the view template path */
                if (!utils.diskio.existsSync(config.paths.templates + '/view.' + config.viewEngine))
                    throw new Error('View template could not be found. Template required to auto generate view.');

                /* get the view template */
                template = utils.diskio.readSync(config.paths.templates + '/view_' +  viewType + '.' + config.viewEngine);

                /* write the template */
                utils.diskio.writeSync(viewsPath + options.view + viewsExt, template);
            }

        }
        
        /* iterates models and returns models with keyValues property */
        function extendModel(models) {
            var tmpModels = JSON.parse(JSON.stringify(models));
            if (_.isArray(models)) {         
                var tmpArr = []; 
                _.forEach(tmpModels, function (model) {
                    model = trimModel(model) || {};                   
                    model.keyValues = pairModel(model);
                    tmpArr.push(model);
                });
                tmpModels = tmpArr;
            } else {
                tmpModels = trimModel(tmpModels);
                tmpModels.keyValues = pairModel(tmpModels);
            }
            return tmpModels;
        }
        
        /* 
        * Iterates object and creates array of objects
        * containing k/v pairs
        */
        function pairModel(model) {            
            var arr = [], xProp;
            _.forEach(model, function (value, key) {
                xProp = xModel.properties[key];
                if (!_.isArray(value)) arr.push({ key: key, value: value || '', display: xProp.display, typeString: xProp.typeString });
                else arr.push({ key: key, value: pairModel(value), display: xProp.display, typeString: xProp.typeString });
            });
            return arr;
        }

        /* trims properties that should not be returned via REST or rendered model */
        function trimModel(model) {   
           return utils.helpers.trimModel(model, xModel);
        }
        
        /* get number of pages for filtered data */
        function pageCount(count, limit) {
            if (limit === 0 || limit === undefined) return 1;
            var pages;
            pages = Math.floor((count + limit - 1) / limit);
            pages = pages === 0 ? 1 : pages;
            return pages;
        }

        /* 
        * Hacky way to build out url's so
        * jquery/js doesn't have to be used in 
        * auto generated views.
        */
        function prepUrl(u) {
            u = decodeURIComponent(u);
            var uObj, orderBy;
            uObj = url.parse(u, true);
            uObj.query = uObj.query || {};
            orderBy = uObj.query.order || 'id ASC';
            orderBy = orderBy.split(' ');
            if (u.indexOf('?') !== -1) {
                if (u.indexOf('order=') === -1)
                    u += '&order=id';
                if (u.indexOf('page=') === -1)
                    u += '&page=1';
                u = u.replace(/ (ASC|DESC)/i, '');
                lastOrder = orderBy[0];
                lastDir = orderBy[1];
            } else {
                u += '?order=id ASC&page=1';
                lastOrder = 'id';
                lastDir = 'ASC';
            }
            return u;
        }
        
        /* 
         * Manually filter model using predicate or
         * generated statement based on search keys passed.   
         */
        function filterPredicate(model) {
            
            var keys = query.split(' ');

            if (config.adapters.active === 'memory') {
                var valid, regex;
                valid = false;
                _.forEach(model, function (value) {
                    if (valid) return;
                    _.forEach(keys, function (v) {
                        if (valid) return;
                        regex = new RegExp('.*' + v + '.*', 'gi');
                        valid = regex.test(value);
                    });
                });
                return valid;
            }

            if (config.adapters.active === 'mongodb') {

                var orObj = {},
                    orArr = [];
                _.forEach(model, function (value, key) {
                    _.forEach(keys, function (v) {
                        var filterObj = {},
                            regex = new RegExp('.*' + v + '.*', 'gi');
                        filterObj[key] = regex;
                        orArr.push(filterObj);
                    });
                });

                orObj.$or = orArr;
                return orObj;
            }

            if (config.adapters.active === 'sqlite3' || config.adapters.active === 'mysql' ) {

                var orObj = {},
                    orArr = [];
                _.forEach(model, function (value, key) {   
                    _.forEach(keys, function (v) {      
                        var tmpObj = {};
                        v = config.adapters.active === 'sqlite3' ? '%' + v + '%' : '"%' + v + '%"'                     
                        tmpObj[key] = { $like: v };               
                        orArr.push(tmpObj);
                    });
                });
  
                orObj.$or = orArr;
                return orObj;     
            }

            if (config.adapters.active === 'cradle') {

                /* we'll have this running soon */
                throw new NotImplementedError('Adapter Cradle for CouchDB is still under development for auto routes.');
            }
            
            /* 
            * NOTE: Redis adapter does NOT offer complex filtering
            * only a single index will be filtered on without using "LIKE"
            * type operator as other adapters do.
            */
            if (config.adapters.active === 'redis') {

                var filterObj = {}
                  , hasIndex = false;
                _.forEach(model, function (value, key) {
                    if (hasIndex) return;
                    /* only add indexed properties */
                    if (value.index) {
                        filterObj[key] = query;
                        hasIndex = true;
                    }
                });

                /* if we hit here not index in model can't query */
                if (!hasIndex) throw new NullObjectError('Index Not Found. Make sure that the ' + xModel.tableName + ' has one of its properties been set as "index: true".')
                return filterObj;             
            }
        }

        function routePrefix(u) {
            var arr = u.split('/');
            return arr[0] + '/' + (arr[1] || '');
        }

 
        return {
            /**
           * Returns all models in collection.
           * @function all
           * @param {object} req - the http request.
           * @param {object} res - the http response.
           * @param {object} next - next.
           * @augments call
           */
            all: function all(req, res, next) {
                
                /* not needed if REST */
                if (!req.REST) {
                    var countFilter = { where: {} };

                    /* check for search query */
                    if (req.filter.q) {
                        query = req.filter.q;
                        delete req.filter.skip;
                        delete req.filter.q;
                        req.filter.page = req.filter.page || 1;
                        req.filter.where = filterPredicate;
                        countFilter.where = filterPredicate;
                    } else {
                        query = '';
                    }
                    
                    Model.all(countFilter, function (err, models) {
                        if (err || !models) {
                            if (err) next(new DatabaseError(err))
                            else next(new NullObjectError(notFound));                                           
                        } else {

                            var count, page, pages, pageOptions;
                            count = models.length;
                            page = req.filter.page || 1;

                            /* check page size and limit */
                            if (!req.filter.limit && xModel.pageSize !== 0)
                                req.filter.limit = xModel.pageSize;
                            if (req.filter.page)
                                if (page > 1)
                                    req.filter.skip = req.filter.limit * (page - 1);

                            /* get page count */
                            pages = pageCount(count, req.filter.limit);

                            if (count === 0) {
                                pages = 0;
                                page = 0;
                            }

                            /* create object to hold all page properties */
                            pageOptions = {
                                query: query,
                                showId: false,
                                pager: {
                                    page: page,
                                    pages: pages,
                                    count: count
                                },
                                lastOrder: lastOrder,
                                lastDir: lastDir,
                                urlQuery: prepUrl(req.url),
                                urlPrefix: routePrefix(req.url)              
                            };
                            
                            /* get the models */
                            Model.all(req.filter, function (err, models) {
                                if (err) {
                                    next(err);
                                    return false;
                                }
                                res.render(options.view, { model: extendModel(models), options: pageOptions });
                            });
                        }

                    });
                    
                } else {
                    /* get the models */
                    Model.all(req.filter, function (err, models) {
                        if (err) {
                            next(new DatabaseError(err));
                            return false;
                        }
                        res.json(200, trimModel(models));
                    });
                }
            }

            /**
            * Finds model by id.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            find: function find(req, res, next) {
                if (req.method.toLowerCase() !== 'get') {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }
                var id = req.params.id;
                if (!id) {
                    next(new NullObjectError(nullObject));
                    return false;
                } 
                Model.findOne({ where: { id: id } }, function (err, model) {                   
                    if (err || !model) {
                        if (err) next(new DatabaseError(err))
                        else next(new NullObjectError(notFound));
                        return false; 
                    }
                    var pageOptions = {
                        urlPrefix: routePrefix(req.url)
                    }
             
                    if (req.REST)
                        res.json(200, trimModel(model));
                    else
                        res.render(options.view, { model: extendModel(model), options: pageOptions });

                });
            }

            /**
            * Saves model.
            * @function update
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            update: function update(req, res, next) {
                if (!_.contains(['put', 'post'], req.method.toLowerCase())) {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }
                if (!req.body) {
                    next(new NullObjectError(nullObject));
                    return false;
                }

                var id = req.params.id || req.body.id || null,
                    data = req.body;
           
                Model.findOne({ where: { id: id } }, function findOne(err, model) {
                    if (err || !model) {
                        if (err) next(new DatabaseError(err))
                        else next(new NullObjectError(nullObject));
                        return false;
                    }

                    if (data.id) delete data.id;

                    if (config.autoRoute.pruneOnSave) {
                        _.forEach(data, function (val, prop) {
                            if (!val || /^\s*$/.test(val)) delete data[prop];
                            if (prop === 'created' || prop === 'modified') delete data[prop];
                        });
                    }                                   
              
                    _.merge(model, data);          
             
                    model.isValid(function (valid) {
                        if (!valid) {
                            var err = model.errors;
                            next(new DatabaseValidationError(null, err));                            
                        } else {
                            model.save(function save(err) {
                                if (err) {
                                    next(new DatabaseError(err));
                                    return false;
                                }
                                if (req.REST) {
                                    res.json(200, trimModel(model));
                                } else {
                                    res.redirect(options.redirect + '/' + id);  //res.render(options.view, { model: model });
                                }
                            });

                        }
                    });             
                });
                
            }

            /**
            * Creates a new model.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            create: function create(req, res, next) {             
                if (!req.REST && req.method.toLowerCase() === 'get') {
                    var model = extendModel(xModel.properties),
                        pageOptions = {
                            urlPrefix: routePrefix(req.url)
                        }
                    res.render(options.view, { model: model, options: pageOptions });
                } else {

                    var data = req.body || null
                      , model;
                   if (!data) next(new NullObjectError(nullObject));

                   if (config.autoRoute.pruneOnSave) {
                       _.forEach(data, function (val, prop) {
                           if (!val || /^\s*$/.test(val)) delete data[prop];
                           if (prop === 'created' || prop === 'modified') delete data[prop];
                       });
                   }           
                   model = new Model(data);                 
                    /* check if model is valid */
                    model.isValid(function (valid) {              
                        if (!valid) {
                            var err = model.errors;                        
                            next(new DatabaseValidationError(null, err));
                        } else {                      
                            model.save(function (err, model) {                          
                                if (err || !model) {                                   
                                    if (err) next(new DatabaseError(err))
                                    else next(new NullObjectError(notFound));
                                    return false;
                                }                       
                                if (req.REST)
                                    res.json(200, trimModel(model));
                                else
                                    res.redirect(options.redirect);
                            });
                        }
                    });                  
                }
            }

            /**
            * Destroys or deletes a model by id.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            destroy: function destroy(req, res, next) {
                if (!_.contains(['delete', 'post', 'del'], req.method.toLowerCase())) {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }    
                req.body = req.body || {};
                var id = req.params.id || req.body.id || null;
                if (!id) {
                    next(new NullObjectError(nullObject));
                    return false;
                }
                Model.findOne({ where: { id: id } }, function (err, model) {
                    if (err || !model) {
                        if (err) next(new DatabaseError(err))
                        else next(new NullObjectError(notFound));
                        return false;
                    }
                    model.destroy(function (err) {
                        if (err) {
                            next(new DatabaseError(err));
                            return false;
                        }
                        if (req.REST) 
                            res.json(200, 'Id: ' + id + ' was successfully deleted.');
                        else 
                            res.redirect(options.redirect);  //res.render(options.view, { message: 'id: ' + id + ' was successfully deleted.' });
                        
                    });
                });
            }

            /**
            * Checks if a model exists by id.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            exists: function exists(req, res, next) {
                if (!req.REST || req.method.toLowerCase() !== 'get') {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }
                Model.exists(req.params.id, function (err, exists) {
                    if (err) {
                        next(new DatabaseError(err));
                        return false;
                    }
                    res.json(200, { exists: exists });
                });
            }

            /**
            * Counts models in collection.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            count: function count(req, res, next) {
                if (!req.REST || req.method.toLowerCase() !== 'get') {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }
                Model.count(req.filter.where, function (err, count) {
                    if (err) {
                        next(new DatabaseError(err));
                        return false;
                    }
                    res.json(200, { count: count || 0 });
                });

            }

            /**
            * Updates properties for a model by id.
            * @function find
            * @param {object} req - the http request.
            * @param {object} res - the http response.
            * @param {object} next - next.
            * @augments call
            */,
            updateAttributes: function updateAttributes(req, res, next) {
                if (!req.REST || !_.contains([ 'put', 'post'], req.method.toLowerCase())) {
                    next(new NotImplementedError('Invalid operation for the requested method.'));
                    return false;
                }

                var data = req.body || null;

                if (!data || !data.id) {
                    next(new NullObjectError(nullObject));
                    return false;
                }

                if (config.autoRoute.pruneOnSave) {
                    _.forEach(data, function (val, prop) {
                        if (!val || /^\s*$/.test(val)) delete data[prop];                    
                    });
                }

                Model.findOne({ where: { id: id } }, function (err, model) {
                    if (err || !model) {
                        if (err) next(new DatabaseError(err))
                        else next(new NullObjectError(notFound));
                        return false;
                    }
                    model.isValid(function (valid) {
                        if (!valid) {
                            var err = model.errors;
                            next(new DatabaseValidationError(null, err));
                        } else {
                            model.updateAttributes(data, function (err, model) {
                                if (err) {
                                    next(new DatabaseError(err));
                                    return false;
                                }
                                res.json(200, trimModel(model));
                            });
                        }

                    });
             
                });
            }
        };

    };




};
