var Schema = require('jugglingdb').Schema
  , path = require('path');

/**
* Builds models into routes, combines user routes.
*
* TODO refactor this to use async too complex now.
*
* @module init
* @param {object} xipframe - loaded xipframe instance.
*/
module.exports = function init(xipframe) {

    'use strict';

    var utils = xipframe.utils,
        _ = utils._,
        config = xipframe.config,
        supported = config.adapters.supported,
        adapterName = config.adapters.active,
        routes = {},
        definedModel = {},
        userRoutes = {},
        tmpArr = [],
        tmpObj = {},
        dbpath,
        adapter,
        schema;

    adapter = config.adapters[adapterName];

    /* create model schemas and add to xipframe */
    if (!_.contains(supported, adapterName))
        throw new Error(adapterName + ' is not supported. Please enable one of the following: ' + supported.join(', '));

    /* require lib if not included in jugglingdb */
    if (!_.contains(config.adapters.noRequire, adapterName)) {
        if (!adapter.module) throw new Error(adapterName + ' could not be located or is missing required module name.');
        else require(adapter.module);
    }

    /* if sqlite3 check the path */
    if (adapterName === 'sqlite3') {
        adapter.database = config.paths.app + adapter.database;
        if (!utils.diskio.existsSync(config.paths.app + '/db'))
            utils.diskio.mkdir(config.paths.app + '/db');
    }

    /* define new schema */
    schema = new Schema(adapterName, adapter);

    function defineModel() {

        /* define models. */
        _.forEach(xipframe.models, function (model, key) {

            if (key === 'extend') return;

            /* setup model tablename and page size */
            model.tableName = model.tableName || key.replace(/Model/i, '');

            /* create the schema */
            if (!_.has(xipframe.schema.models, model.tableName)) {

                /* merge in global properties */
                if (xipframe.models.extend && xipframe.models.extend.properties) model.properties = _.merge({}, model.properties, xipframe.models.extend.properties);

                /* get the model page size */
                if (model.pageSize && model.pageSize !== 0)
                    model.pageSize = model.pageSize || config.pageSize;
                else
                    model.pageSize = config.pageSize;

                if (!model.properties.id) model.properties.id = { type: Number, order: 9999, index: true };

                /* set some needed defaults */
                _.forEach(model.properties, function (value, key) {

                    tmpArr = [];
                    tmpObj = {};

                    /* get the type and create lowercase string representation */
                    var typeToString;
                    if (value.type === 'Text' || value.type === 'JSON') {
                        typeToString = value.type.toLowerCase();
                        value.type = value.type === 'Text' ? Schema.Text : Schema.JSON;
                    } else {
                        typeToString = utils.helpers.getType(new value.type);
                        //value.type = new value.type;
                    }

                    /* make sure numbers have a default or they'll be missing when enumerating objects */
                    if (key !== 'id') {
                        if (typeToString === 'number') value.default = value.default || 0;

                        /* makes sure properties have a default so they are not undefined */
                        if (config.autoRoute.undefinedIsNull) {
                            if (_.contains(['string', 'text', 'json', 'object', 'array'], typeToString)) value.default = value.default || null;
                            if (typeToString === 'boolean') value.default = (value.default === true || value.default === false ? value.default : null);
                        }
                    }

                    /* set a few defaults */
                    value.typeString = typeToString;
                    value.order = key === 'modified' || key === 'created' ? value.order || 9999 : value.order || 999;

                    /* just sets camel case to title case or if you're picky its really Start Case */
                    if (!value.display && config.autoRoute.headerToTitle) {
                        var str = key.match(/[A-Z]?[a-z]+|[0-9]+/g).join(' ');
                        value.display = str.charAt(0).toUpperCase() + str.slice(1);
                    } else {
                        value.display = value.display || key;
                    }

                    value.hidden = value.hidden || false;

                    /* we'll use this in a minute to reorder the properties */
                    tmpArr.push({ key: key, order: value.order });

                });

                /* sorts the array by multiple properties */
                var sortArr = ['order'];
                if (config.autoRoute.alphabetize) sort.push('key');
                utils.helpers.sortArray(tmpArr, sortArr);
                _.forEach(tmpArr, function (value) {
                    var prop = value.key;
                    tmpObj[prop] = {};
                });

                /* finally merge the ordered properties. */
                model.properties = _.merge({}, tmpObj, model.properties);

                /* define the model */
                definedModel = schema.define(model.tableName, model.properties);

                /* add global life-cycle callbacks */
                if (xipframe.models.extend && xipframe.models.extend.callbacks) _.merge(definedModel, xipframe.models.extend.callbacks, model.callbacks);

                /* migration for sql */
                if (adapterName === 'mysql' || adapterName === 'sqlite3' && adapter.scheme) {
                    if (adapter.scheme) {
                        if (adapter.scheme === 'autoupdate') schema.autoupdate(function () {
                            modelSeeds(model, definedModel);
                        });
                        if (adapter.scheme === 'automigrate' && adapterName === 'mysql') schema.automigrate(function () {
                            modelSeeds(model, definedModel);
                        });
                    }
                } else {
                    modelSeeds(model, definedModel);
                }

            } else {
                xipframe.log.warn('[DuplicateModel] duplicate model detected. Offending model will be ignored.');
            }

        });

        modelExtend();
    }

    /* seed in models if in development mode */
    function modelSeeds(model, definedModel) {

        if (xipframe.config.environment !== 'development' || !xipframe.config.allowSeeds
            || !model.seeds.enabled || !model.seeds.data || !_.isArray(model.seeds.data)) return;

        console.log(' > Seeding model data for ' + model.tableName);

        definedModel.destroyAll(function (err) {
            if (err) {
                xipframe.log.warn('[' + err.name || 'Unknown Error' + '] ' + err.message);
            } else {
                _.forEach(model.seeds.data, function (seed) {
                    var seedModel = new definedModel(seed);
                    seedModel.save(function (err) {
                        if (err) xipframe.log.warn('[' + err.name || 'Unknown Error' + '] ' + err.message);
                    });
                });
            }
        });
    }

    /* extend the model with relationships and validations */
    function modelExtend() {
        _.forEach(schema.models, function (model, key) {
            var xModel = xipframe.models[key + 'Model'];
            if (xModel.extend) xModel.extend(model, schema);
        });
        buildRoutes();
    }

    /* build our routes based on models and user defined routes */
    function buildRoutes() {

        /* update xipframe schema object */
        xipframe.schema = schema;

        /* merge user routes */
        _.forEach(xipframe.routes, function (routes) {
            _.forEach(routes, function (route, key) {
                userRoutes[key] = route;
            });
        });


        var modelRoutes = {}; //userRoutes = xipframe.config.routes || null          

        /* normalize user routes */
        if (userRoutes) {

            _.forEach(userRoutes, function (route, path) {

                var disabled = route.disabled || false
                  , redirect = route.redirect || false;

                path = path.toLowerCase();               

                if (redirect && !disabled && !_.has(xipframe.express.redirects, route.redirect))
                    xipframe.express.redirects[path] = route.redirect;

                if (!disabled && !redirect && route.controller && route.action) {

                    route.method = route.method || 'get';

                    /* check if route contains method if so set it to the route method and strip it out */
                    _.forEach(['get', 'post', 'put', 'del'], function (m) {
                        if (path.indexOf(m) === 0) {
                            route.method = m;
                            path = path.replace(m + ' ', '');
                        }
                    });

                    path = route.method + ' ' + path;

                    /* load the controller and action */
                    var controller, action;
                    action = null;
                    controller = utils.helpers.getPropertyByName(xipframe.controllers, route.controller, [/Controller/i]);

                    if (controller) {
                        action = utils.helpers.getPropertyByName(controller, route.action);
                    }

                    /* if controller and route exits add the route */
                    if (controller && action) {
                        routes[path] = {
                            action: action,
                            path: path,
                            method: route.method,
                            actionName: route.action.toLowerCase(),
                            controller: route.controller.toLowerCase()
                        };
                    }
                } else {
                    // invalid route controller or route action specified.  
                    if (!disabled && !redirect)
                        xipframe.log.warn('[InvalidRoute] ' + path + '" is corrupt or invalid and could not be added. Verify a valid action exists.');
                }
            });
        }


        /* build autoRutes for models */
        if (config.autoRoute.enabled) {

            var methods = []
            , resultRoute = {};

            _.forEach(xipframe.schema.models, function (model, name) {
           
                if (!config.autoRoute.getPostOnly)
                    methods = ['get', 'get:id', 'put', 'post', 'del', 'get:exists', 'get:count', 'post:update', 'post:delete', 'post:attributes', 'post:showall', 'post:show', 'get:create'];
                else
                    methods = ['get', 'get:id', 'get:create', 'get:exists', 'get:count', 'post', 'post:update', 'post:delete', 'post:attributes', 'post:showall', 'post:show'];
                    //methods = ['get', 'get:id', 'get:create', 'get:exists', 'get:count', 'post', 'post:update', 'post:delete', 'post:attributes', 'post:showall', 'post:show', 'put', 'put:attributes', 'del'];
                _.forEach(methods, function (method) {           
                    resultRoute = buildAutoRoute(name, method, model);
                    modelRoutes[resultRoute.path] = resultRoute;
                });

            });


            /* merge in the routes, userRoutes overwrite auto generated routes */
            routes = _.merge({}, modelRoutes, routes);
       
        }

    }

    /* 
    * creates auto routes 
    *
    * TODO: need to go through this as there might be some redundancy.
    */
    function buildAutoRoute(name, method, model) {

        method = method.toLowerCase();
        var actions = require('./actions')(xipframe),
            route = {},
            tmpMethod = method.split(':')[0].toLowerCase(),
            prefix = tmpMethod + ' /', options = {},
            xModel;

        tmpMethod = tmpMethod === 'delete' ? 'del' : tmpMethod;

        /* plucks out the original model */
        xModel = _.find(xipframe.models, function (m, key) {
            var regex = new RegExp(name + 'Model');
            return regex.test(key);
        });

        options.xModel = xModel;
        options.model = model;
        name = name.toLowerCase();
        route.controller = name;

        if (_.contains(['post', 'put', 'delete', 'del', 'get:id', 'get:exists', 'get:count',
            'put:attributes', 'post:update', 'post:delete', 'post:attributes', 'post:showall',
            'post:show', 'get:create' ], method)) {

            route.method = tmpMethod;

            switch (method) {

                /* get methods */
                case 'get:count':
                    route.path = prefix + name + '/count';
                    //options.view = name + '/' + 'count';
                    route.action = actions(options).count;
                    route.actionName = 'count';
                    break;
                case 'get:id':
                    route.path = prefix + name + '/show/:id';
                    options.view = name + '/' + 'show';
                    route.action = actions(options).find;
                    route.actionName = 'find';
                    break;
                case 'get:exists':
                    route.path = prefix + name + '/exists/:id';
                    //options.view = name + '/' + 'exists';
                    route.action = actions(options).exists;
                    route.actionName = 'exists';
                    break;
                case 'get:create':
                    route.path = prefix + name + '/create';
                    options.view = name + '/' + 'create';
                    route.action = actions(options).create;
                    route.actionName = 'create';
                    break;

                    /* post methods */
                case 'post':
                    route.path = prefix + name + '/create';
                    //options.view = name + '/' + 'create';
                    options.redirect = '/' + name + '/' + 'show'; /* redirect after post */
                    route.action = actions(options).create;
                    route.actionName = 'create';
                    break;
                case 'post:update':
                    route.path = prefix + name + '/update/:id?';
                    //options.view = name + '/' + 'update';
                    options.redirect = '/' + name + '/' + 'show'; /* redirect after post */
                    route.action = actions(options).update;
                    route.actionName = 'update';
                    break;
                case 'post:delete':
                    route.path = prefix + name + '/destroy/:id?';
                    //options.view = name + '/' + 'destroy';
                    options.redirect = '/' + name + '/' + 'show'; /* redirect after post */
                    route.action = actions(options).destroy;
                    route.actionName = 'destroy';
                    break;
                case 'post:attributes':
                    route.path = prefix + name + '/updateAttributes';
                    //options.view = name + '/' + 'updateAttributes';
                    route.action = actions(options).updateAttributes;
                    route.actionName = 'updateAttributes';
                    break;
                case 'post:showall':
                    route.path = prefix + name + '/show';
                    options.view = name.toLowerCase() + '/' + 'showall';
                    route.action = actions(options).all;
                    route.actionName = 'all';
                    break;
                case 'post:show':
                    route.path = prefix + name + '/show/:id';
                    options.view = name + '/' + 'show';
                    route.action = actions(options).update;
                    options.redirect = '/' + name + '/' + 'show'; /* redirect after post */
                    route.actionName = 'update';
                    break;


                    /* put methods */
                case 'put':
                    route.path = prefix + name + '/update/:id?';
                    //options.view = name + '/' + 'update';
                    route.action = actions(options).update;
                    route.actionName = 'update';
                    break;
                case 'put:attributes':
                    route.path = prefix + name + '/updateAttributes/:id';
                    //options.view = name + '/' + 'updateAttributes';
                    route.action = actions(options).updateAttributes;
                    route.actionName = 'updateAttributes';
                    break;


                    /* delete methods */
                case 'del':
                    route.path = prefix + name + '/destroy/:id?';
                    //options.view = name + '/' + 'destroy';
                    route.action = actions(options).destroy;
                    route.actionName = 'destroy';
                    break;
                case 'delete':
                    route.path = prefix + name + '/destroy/:id?';
                    //options.view = name + '/' + 'destroy';
                    route.action = actions(options).destroy;
                    route.actionName = 'destroy';
                    break;

            }

        } else {
            route.path = prefix + name + '/show';
            route.method = method;
            options.view = name.toLowerCase() + '/' + 'showall';
            route.action = actions(options).all;
            route.actionName = 'all';
        }
        return route;
    }

    defineModel();

    /* add the routes to the instance */
    xipframe.routes = routes;

};
