var defaults = require('../configuration').defaults
   , Chance = require('chance')
   , ejs = require('ejs');

/**
* Installs a new Xipframe application.
* @module install
* @param {object} xipframe - the current xipframe instance.
* @param {object} userConfig - the user defined configuration.
*/
module.exports = function install(xipframe, userConfig) {

    'use strict';

    var sessionId, chance, removeSeeds, app, isInstalled, rootPath,
        pkgData, tmpConfig, progressBar, log, utils, _, config,
        InstallError;

    config = xipframe.config;
    utils = xipframe.utils;
    _ = utils._;
    InstallError = xipframe.errors.InstallError;

    /* should never hit here but just in case */
    if (!userConfig) throw new InstallError('Invalid or corrupt configuration, exiting...');

    removeSeeds = [];
    chance = new Chance();

    console.log('');
    console.log(' ###############################################################');
    console.log(' > Installing Xipframe Application.');
    console.log(' ###############################################################');
    console.log('');

    /* If config load has value load the config file passed */
    if (userConfig.config && userConfig.config.load) {
        userConfig = utils.package.get(userConfig.load);
        if (!utils.isJSON(userConfig)) throw new InstallError('Specified configuration is invalid.');
    }

    /* create a secret for the sessions */
    userConfig.session.secret = chance.string({ length: 32 });

    /* merge the config and get defaults */
    defaults(xipframe, userConfig);

    /* set var to check if app is already installed */
    isInstalled = utils.appdate.isDuplicate(config.apps, config.appName).exists;

    /* set files for removal based on view engine selected. */
    _.forEach(config.viewEngines.supported, function (ve) {
        if (config.viewEngine !== ve)  removeSeeds.push('.+\.' + ve);
    });

    /* Check if install path already exists */
    if (utils.diskio.existsSync(config.paths.app) || isInstalled) 
        throw new InstallError('Directory <' +  config.paths.app + '> not empty or Application installed. Backup data then run "xipframe uninstall ' + config.appName + '"');
    

    console.log(' > Copying file and folder structure...this may take a moment.');

    /* 
    * Recursively seed/copy files and folders 
    * TODO: check for missing dependencies etc. 
    */ 
    //progressBar = utils.helpers.showProgress();
    utils.diskio.copyRecursive(config.paths.seeds, config.paths.app, function () {
        utils.diskio.copyRecursive(config.paths.root + '/node_modules/grunt', config.paths.app + '/node_modules/grunt', function () {
            //utils.helpers.stopProgress(progressBar);               
            seedData();           
        });              
    });

    /* seeds in data  into app config files from our install params and default params */
    function seedData() {   

        console.log(' > Seeding initial data.');

        /* if less create the initial file using lessFilename property.*/
        if (config.less.enabled && !config.less.lessFilename) 
            utils.diskio.writeSync(config.paths.less + '/' + config.less.lessFilename + '.less', '');

        /* add escaped root path */
        /* kinda hacky but only need this one path. */
        config.paths.rootEscaped = config.paths.root.replace(/\\/g, "\\\\");   
          
        /* render files with seed data */
        for (var i = 0, seed; seed = config.seeds[i]; i++) {
            var dir = seed.dir;
            for (var n = 0, file; file = seed.files[n]; n++) {
                var result = utils.helpers.render(dir + '/' + file, config);
                utils.diskio.writeSync(dir + '/' + file, result);
            }
        }   
       
        cleanUp();
    }

    function cleanUp() {

        console.log(' > Clean up...removing unnecessary files and folders.');

        /* delete Angular see if not used. */
        if (!config.angular.enabled) utils.diskio.removeSync(config.paths.angular);

        /* delete Less seed if not used. */
        if (!config.less.enabled) utils.diskio.removeSync(config.paths.less);
          
        /* remove examples if not requested. */
        if (!config.examples) utils.diskio.removeSync(config.paths.examples);

        /* add package.json */
        pkgData = { name: config.appName, dependencies: { xipframe: config.version }, scripts: { start: config.start } };
        utils.package.createSync(config.paths.app + '/package.json', pkgData);

        /* clone a new instance of our config */
        tmpConfig = _.clone(config, true);
        app = utils.appdate.get(tmpConfig, true);

        /* add the new app to apps and save to apps.json */
        config.apps[config.appName] = app;
        utils.appdate.save(config, true);

        /* delete unused views & configs */
        utils.diskio.removeRecursiveFilter(config.paths.app,
            removeSeeds, function (removed, count) {

                /* show clean up stats */
                console.log(' > Removed ' + count + ' unnecessary files/folders.');

                /* all done  */
                console.log('\n > Application "' + config.appName + '" successfully installed at "' + config.paths.app + '"');

                /* make sure we exit here */
                process.exit(0);

         });

    }    

};
