
/**
* Gets a complete config after merging in application settings with defaults.
* @module defaults
* @param {object} xipframe - the loaded xipframe instance.
* @param {object} userConfig - the loaded user defined settings.
* @returns {object}.
*/
module.exports = function defaults(xipframe, userConfig) {

    'use strict';

    var appPath, basePaths, paths, angularPaths, config, frameworkPaths,
        miscPaths, angularPaths, protectedProperties, rootPath, InitError,
        utils, _, helpers, InitError, overrides, appExists, removeLinker;

    utils = xipframe.utils;
    _ = utils._;
    helpers = utils.helpers;
    InitError = xipframe.errors.InitError;
    overrides = require('./override')(xipframe)();
    appExists = xipframe.config.apps[xipframe.config.appName] ? true : false

    if (!userConfig) throw new InitError('Invalid Xipframe configuration object exiting...');

    /* just don't want to type long namespace */
    appPath = xipframe.config.paths.app || userConfig.paths.app;
    rootPath = xipframe.config.paths.root;

    /* Primary paths */
    paths = {
        app:        appPath
      , client:     appPath + '/client'
      , framework:  appPath + '/framework'
      , configure:  appPath + '/config'
      , assets:     appPath + '/assets'
    };

    /* Framework paths */
    frameworkPaths = {
        controllers:    paths.framework + '/controllers'
      , models:         paths.framework + '/models'
      , modelHooks:     paths.framework + '/models/hooks'
      , middleware:     paths.framework + '/middleware'
      , security:       paths.framework + '/security'
      , examples:       paths.app + '/examples'
      , handlers:       paths.framework + '/handlers'
      , routes:         paths.framework + '/routes'
    };

    /* Misc paths */
    miscPaths = {
        'public':       paths.client + '/public'
      , views:          paths.client + '/views'
      , less:           paths.assets + '/less'
      , angular:        paths.assets + '/angular'
      , angularPublic:  paths.client + '/angular'
      , log:            appPath + '/logs'
      , override:       rootPath + '/override.json'
    };

    /* Paths used in Angular */
    angularPaths = {
        angular_app:            miscPaths.angular + '/app.js'
      , angular_bootstrap:      miscPaths.angular + '/bootstrap.js'
      , angular_controllers:    miscPaths.angular + '/controllers'
      , angular_factories:      miscPaths.angular + '/factories'
      , angular_services:       miscPaths.angular + '/services'
      , angular_filters:        miscPaths.angular + '/filters'
      , angular_directives:     miscPaths.angular + '/directives'
    };

    /* Main configuration */
    config = {
        appName: 'xipframe'  
      , environment: 'development'
      , host: '127.0.0.1'
      , port: 1337
      , viewEngine: 'ejs'
      , examples: true
      , pageSize: 10
      , cluster: false // NOT IMPLEMENTED
      , indexRender: 'index/index'
      , allowSeeds: true
      , grunt: {
          enabled: true,
          watch: true
      }
      , angular: {
          enabled: false,
          enabledBootstrapper: false,
          minify: false,
          bootstrapper: 'bootstrapper.js',
          app: 'app.js'       
      }
      , jquery: {
          enabled: true
      }     
      , less: {
          enabled: false,
          lessFilename: 'stylesless'
      }
      , handlers: {    
          locals: true,
          error: {
              enabled: true,
              redirect: { type: 'auto', path: null }
          },
          fnf: {
              enabled: true,
              redirect: { type: 'auto', path: null }
          },
          security: {
              enabled: true,     
              redirect: { type: 'auto', path: null, message: { 401: 'Not authorized.', 403: 'Access denied.' } }, // type can be 'view path', 'url' or 'auto' if auto uses shared/403 and returns view else if REST returns message with 403.
              sessionProperty: 'authenticated' // e.g. req.session.authenticated            
           }
      }
      , ssl: {
          enabled: false,
          key: null,
          cert: null
      }
      , autoRoute: {
          enabled: true,        
          getPostOnly: false,
          createViews: true,
          headerToTitle: true,
          alphabetize: false,
          undefinedIsNull: true,
          pruneOnSave: true
      }
      , linker: {
          index: null,
          enabled: true,
          files: [],
          templates: {
              jade: {
                  style: 'link(type="text/css", rel="stylesheet", href="%s")',
                  script: 'script(src="%s", type="text/javascript")',
                  scriptOnly: '<script>%s</script>',
                  stylesTag: '// STYLES',
                  stylesEndTag: '// STYLES END',
                  scriptsTag: '// SCRIPTS',
                  scriptsEndTag: '// SCRIPTS END'
              },
              ejs: {
                  style: '<link type="text/css" rel="stylesheet" href="%s" />',
                  script: '<script src="%s" type="text/javascript"></script>',
                  scriptOnly: 'script \n %s',
                  stylesTag: '<!--STYLES-->',
                  stylesEndTag: '<!--STYLES END-->',
                  scriptsTag: '<!--SCRIPTS-->',
                  scriptsEndTag: '<!--SCRIPTS END-->'
              },
              html: {
                  style: '<link type="text/css" rel="stylesheet" href="%s" />',
                  script: '<script src="%s" type="text/javascript"></script>',
                  scriptOnly: 'script \n %s',
                  stylesTag: '<!--STYLES-->',
                  stylesEndTag: '<!--STYLES END-->',
                  scriptsTag: '<!--SCRIPTS-->',
                  scriptsEndTag: '<!--SCRIPTS END-->'
              }
          }
      }
      , log: {
          level: 'info',
          maxsize: 1024 * 1024 * 10,
          handleExceptions: false,
          transports: ['console', 'file'],
          showStack: false
      }
      , paths: {}
      , session: {
          adapter: 'memory'
        , maxAge: 1000 * 60 * 60 * 24
        , key: 'xipframe.sid'
        , httpOnly: true
      }
      , adapters: {
          active: 'memory'
      }
      , io: {
          enabled: true                  
        , adapters: {
            active: 'memory'
        }
        , options: {
            transports: null 
        }
        , extend: null        
      }
      , express: {
          bodyParser: require('express').bodyParser({ limit: 10000000, defer: true, keepExtensions: true })
        , cookieParser: require('express').cookieParser()
        , methodOverride: require('express').methodOverride()
        , csrf: null 
        , extend: null
      },
      globals: {
          _log: 'log',
          _: 'utils._',      
          _diskio: 'utils.diskio',
          _models: 'schema.models',
          _xmodels: 'models',
          _helpers: 'utils.helpers.public',
          _ignores: 'config.staticIgnore',
          _errors: 'errors'
      },
      modelCallbacks: ['afterInitialize', 'beforeCreate', 'afterCreate', 'beforeSave', 'beforeupdate', 'afterUpdate', 'beforeDestroy', 'afterDestroy', 'beforeValidation', 'afterValidation']
    };

    /* Merge all paths */
    config.paths = paths;

    /* protected properties */
    protectedProperties = {
        autoRoute: {
            views: ['show', 'showall', 'create' ]
        },
        adapters: {
            supported: ['memory', 'mongodb', 'redis', 'mysql', 'sqlite3', 'cradle' ],
            supportedSession: ['memory', 'mongodb', 'redis'],
            noRequire: ['memory', 'riak', 'cradle', 'neo4j', 'http']
        },
        viewEngines: {
            supported: [ 'ejs', 'jade', 'html' ]
        },
        seeds: [
            { dir: paths.configure, files: ['app.js', 'adapters.js', 'session.js', 'io.js', 'security.js', 'linker.js'] },
            { dir: paths.app, files: ['Gruntfile.js'] },
            { dir: frameworkPaths.controllers, files: [ 'IndexController.js']}
        ],
        session: {
            supported: ['memory', 'redis', 'mongodb']
        },
        paths: {
            seeds: rootPath + '/seeds'
            , templates: rootPath + '/templates'
        },
        staticIgnore: { paths: ['/img', '/css', '/js', '/angular', '/socket.io'], regex: null },
        fixups: {
            linker: { source: rootPath + '/fixup/node_modules/grunt-scriptlinker/tasks/scriptlinker.js', target: rootPath + '/node_modules/grunt-scriptlinker/tasks/scriptlinker.js', backup: true },
            mongodb: { source: rootPath + '/fixup/node_modules/jugglingdb-mongodb/lib/mongodb.js', target: rootPath + '/node_modules/jugglingdb-mongodb/lib/mongodb.js', backup: true },
            redis: { source: rootPath + '/fixup/node_modules/jugglingdb-redis/lib/redis.js', target: rootPath + '/node_modules/jugglingdb-redis/lib/redis.js', backup: true },
            sqlite3: { source: rootPath + '/fixup/node_modules/jugglingdb-sqlite3/lib/sqlite3.js', target: rootPath + '/node_modules/jugglingdb-sqlite3/lib/sqlite3.js', backup: true },
            mysql: { source: rootPath + '/fixup/node_modules/jugglingdb-mysql/lib/mysql.js', target: rootPath + '/node_modules/jugglingdb-mysql/lib/mysql.js', backup: true },
            memory: { source: rootPath + '/fixup/node_modules/jugglingdb/lib/adapters/memory.js', target: rootPath + '/node_modules/jugglingdb/lib/adapters/memory.js', backup: true },
            relations: { source: rootPath + '/fixup/node_modules/jugglingdb/lib/relations.js', target: rootPath + '/node_modules/jugglingdb/lib/relations.js', backup: true },
            gruntfile: { source: rootPath + '/fixup/app/Gruntfile.js', target: appPath + '/Gruntfile.js', backup: true },
            linker_config: { source: appPath + '/config/linker.js', target: appPath + '/config/linker.js.bak', backup: false },
            multipart: { source: rootPath + '/fixup/node_modules/express/node_modules/connect/lib/middleware/multipart.js', target: rootPath + '/node_modules/express/node_modules/connect/lib/middleware/multipart.js', backup: true },
            limit: { source: rootPath + '/fixup/node_modules/express/node_modules/connect/lib/middleware/limit.js', target: rootPath + '/node_modules/express/node_modules/connect/lib/middleware/limit.js', backup: true }
        }      
    };

    /* merge in the user config */
    _.merge(config, userConfig);

    /* Merge sub paths here to prevent user overrite */
    _.merge(config.paths, frameworkPaths, miscPaths);

    removeLinker = [];

    /* if less link it */
    if (config.less.enabled) {

        /* add to linker if needed */
        if(!helpers.existsByObject(config.linker.files, { type: 'less' }))  
            config.linker.files.push({ type: 'less', src: '/css/' + config.less.lessFilename + '.css', enabled: true });

        /* seed in data structure if missing */
        if (appExists && !utils.diskio.existsSync(config.paths.less)) 
            utils.diskio.copyRecursive(protectedProperties.paths.seeds + '/assets/less', config.paths.less);          
    } else {
        removeLinker.push('less');
    }

    /* check if jQuery is enabled */
    if (config.jquery.enabled) {
        if (!helpers.existsByObject(config.linker.files, { type: 'jquery' }))
            config.linker.files.push({ type: 'jquery', src: 'https://ajax.googleapis.com/ajax/libs/jquery/[REPLACE]/jquery.min.js', replace: '1.10.2', enabled: true });
    } else {
        removeLinker.push('jquery');
    }

    /* check if Angular is enabled */
    if (config.angular.enabled) {

        _.merge(config.paths, angularPaths);

        if (!helpers.existsByObject(config.linker.files, { type: 'angular' })) {
            config.linker.files.push({ type: 'angular', src: 'https://ajax.googleapis.com/ajax/libs/angularjs/[REPLACE]/angular.min.js', replace: '1.0.8', enabled: true });
            config.linker.files.push({ type: 'angular', src: '/' + config.angular.app, enabled: true });         
        }
        if (config.angular.enabledBootstrapper === true) {
            if(!helpers.existsByObject(config.linker.files, { type: 'angular-boot' }))
                config.linker.files.push({ type: 'angular', src: '/' + config.angular.bootstrapper, enabled: true });
        }

        if (appExists && !utils.diskio.existsSync(config.paths.angular)) 
            utils.diskio.copyRecursive(protectedProperties.paths.seeds + '/assets/angular', config.paths.angular);
    } else {
        removeLinker.push('angular');
        removeLinker.push('angular-boot');
    }

    /* add socket.io to linker */
    if (config.io.enabled) {
        if(!helpers.existsByObject(config.linker.files, { type: 'socket' }))
            config.linker.files.push({ type: 'socket', src: '/socket.io/socket.io.js', enabled: true });
    } else {   
        removeLinker.push('socket');
    }

    /* add default linker linkTo path */
    if (!config.linker.index) config.linker.index = '/layout.' + config.viewEngine;

    /* remove invalid linker groups */
    config.linker.files = helpers.removeLinkerByType(config.linker.files, removeLinker);

    /* merge in protected properties */
    _.merge(config, protectedProperties);   

    /* set log path in log config */  
    config.log.path = config.paths.log + '/server.log';

    /* 
    * build up regex ignores, to ignore requests to static files.
    */
    if (config.staticIgnore.paths && config.staticIgnore.paths.length > 0) {
        var arr = config.staticIgnore.paths;
        _.forEach(arr, function (path, key) {
            arr[key] = '^' + path + '/.+$';
        });
        config.staticIgnore.regex = new RegExp('(' + arr.join('|') + ')', 'i');
    }

    _.merge(xipframe.config, config, overrides);

};
