
/**
* Handles errors for domain.
* @module errorHandler
* @param {object} xipframe - the loaded instance of Xipframe.
* @returns {function}
*/
module.exports = function errorHandler(xipframe) {

    'use strict';

    var utils = xipframe.utils
      , errorConfig = xipframe.config.handlers.error
      , log = xipframe.log
      , logconf = xipframe.config.log
      , trace = utils.trace
      , _ = utils._;

      /**
      * Handles errors.
      *
      * @function errorHandlerMiddleware
      * @param {object} err - the current error if any.
      * @param {object} req - the http request.
      * @param {object} res - the http response.
      * @param {function} next - continues request. 
      * @augments errorHandler
      */
    return function errorHandlerMiddleware(err, req, res, next) {

        err = trace.format(err);
        err.name = err.name || 'Unkown Exception';
   
        var metadata = {

            line: err.metadata.line,
            col: err.metadata.col,
            func: err.metadata.function,
            file: err.metadata.file            

        },
        tmpStack = trace.stackToString(err.metadata.stack, '|');
        
        /* make sure we have detected if this is a RESTful request or not.*/
        if (!req.REST || req.REST === 'undefined') {
            req.REST = req.xhr === true;
            if (req.headers['accept']) req.REST = (req.headers['accept'].indexOf('application/json') !== -1);
            if (req.headers['content-type']) req.REST = (req.headers['content-type'].indexOf('application/json') !== -1);
        }           

        /* 
        * check if message is object and is returning to error page. 
        * if yes improve the formatting. 
        */       
        if (err.codes) {            
            var tmpMsg = [];
            _.forEach(err.codes, function (obj, key) {           
                var isArray = _.isArray(obj);         
                tmpMsg.push(key + ': ' + (isArray ? obj.join(', ') : obj ));
            });
            err.message = tmpMsg.join(', ') || 'Unknown validation errors.';
        }

        /* set some vars thant can be handy when dealing with errors */
        err.showStack = logconf.showStack;
        err.environment = xipframe.config.environment;

        /* remove the stack trace unless specified to show */
        if (logconf.showStack) metadata.stack = tmpStack;
        else delete err.metadata.stack;

        /*
        * Error Properties:
        * err.message, err.name, err.metadata.stack, err.metadata.function, 
        * err.metadata.file, err.metadata.line, err.metadata.col.    
        *
        */
        if (errorConfig.enabled || !xipframe.userHandlers.error && err) {

            /* log the error */
            log.error('[' + err.name + '] ' + err.message, metadata);

            /* build up error if rendered in view */
            if (!req.REST) {
                if (logconf.showStack) {
                    err.metadata.stack = err.metadata.stack.replace(/\|/gi, '<br/>');
                    err.metadata.stack = encodeURIComponent(err.metadata.stack)
                }

                if (xipframe.config.environment !== 'development') {
                    err.metadata.line = '';
                    err.metadata.col = '';
                    err.metadata.func = '';
                    err.metadata.file = '';
                    err.metadata.stack = '';
                }

                /* adds a local variable for the last error */
                if(res.locals && res.locals.system) res.locals.system.lastError = err;
            }         
            
            if (errorConfig.redirect.type === 'auto' || req.REST) {      
                if(req.REST) res.json(500, err);
                else res.render('shared/error', { err: err, title: res.locals.title || 'Error 500' });
            } else {
                if (errorConfig.redirect.type === 'view') {
                    if (errorConfig.redirect.path) res.render(errorConfig.redirect.path, { err: err, title: res.locals.title || 'Error 500' });
                    else throw new Error('Invalid view path for error handler.');
                } else if (errorConfig.redirect.type === 'url') {
                    if (errorConfig.redirect.path) res.redirect(errorConfig.redirect.path);
                    else throw new Error('Invalid redirect path for error handler.');
                } else {
                    res.render('shared/error', { err: err, title: res.locals.title || 'Error 500' });
                }            
            }        
  
        } else {

            if (err)
                next(err); 
            else
                next(); 

        }

    };    

}; 
