
var url = require('url');

/**
* Handles the security policy.
* @module securityHandler
* @param {object} xipframe - the loaded instance of xipframe.
*/
module.exports = function securityHandler(xipframe) {

    'use strict';

    var config = xipframe.config
      , utils = xipframe.utils
      , _ = utils._
      , routes = xipframe.routes
      , authConfig = xipframe.config.handlers.security
      , sessionProperty = authConfig.sessionProperty;


    /**
     * Validates and applies the security policy.
     *
     * @function securityHandlerMiddleware   
     * @param {object} req - the http request.
     * @param {object} res - the http response.
     * @param {function} next - continues request.
     * @param {function|string|boolean} policy - the policy to apply.
     * @augments securityHandler
     */
    return function securityHandlerMiddleware(req, res, next) {


        if (!config.staticIgnore.regex.test(req.url)) {

            var policy = xipframe.config.security['*'] || false
           , args = Array.prototype.slice.call(arguments)
           , reqUrl = url.parse(req.url)
           , method = req.method.toLowerCase()
           , appRoutes = req.app.routes[method]
           , route
           , routePath
           , controllerPolicies;

            route = _.find(appRoutes, function (route) {
                return route.regexp.test(req.url) === true;
            });

            method = method === 'delete' ? 'del' : method;
 
            if (xipframe.config.handlers.security.enabled && route !== undefined && route !== null && route.path) {

                /* lookup the framework route, controller and security policies */
                route = utils.helpers.getPropertyByName(routes, method + ' ' + route.path);

                if (!route || !route.controller) throw new Error('Invalid route or controller.');

                if (!handled()) {

                    /* get the security policy for the controller */
                    controllerPolicies = utils.helpers.getPropertyByName(config.security, route.controller, [/Controller/i]);               

                    if (controllerPolicies)
                        policy = controllerPolicies['*'] === undefined ? true : controllerPolicies['*'];

                    if (!handled()) {

                        policy = utils.helpers.getPropertyByName(controllerPolicies, route.actionName, false, false, true);                       

                        if (!handled()) {

                            next();
                        }
                    }
                }

            } else {

                next();

            }
        
        } else {

            next();

        }       


        function handled() {

            /* if string lookup the handle and call/apply args */
            if (!_.isBoolean(policy)) {        

                /* user entered a callback function directly or specified a callback handle */
                policy = utils.helpers.lookupHandle(xipframe.security, policy, false);

                if (policy !== false) {

                    policy.apply(this, args);

                } else {

                    deny();

                }
                
                /* return true regardless sinced handled */
                return true;
      
            }

            if (policy === false) {

                /* policy requires user to be authenticated */
                if (!req.session[sessionProperty]){

                    deny(401);

                    return true;

                }                

            }

            /* if we get here just return false need to keep checking policies. */
            return false;

        }

        function deny(status) {

            var msg = authConfig.redirect.message[status]
              , path = authConfig.redirect.path  || ('shared/' + status);

            if (authConfig.redirect.type === 'auto' || req.REST) {
                if (req.REST) res.json(status, msg);
                else res.render('shared/' + status, { url: req.url, title: status + ' - ' +  msg, message: msg });
            } else {
                if (authConfig.redirect.type === 'view') {
                    if (authConfig.redirect.path) res.render(authConfig.redirect.path, { url: req.url, title: status + ' - ' + msg, message: msg });
                    else throw new Error('Invalid view path for security handler.');
                } else if  (authConfig.redirect.type === 'url') {
                    if (authConfig.redirect.path) res.redirect(authConfig.redirect.path);
                    else throw new Error('Invalid redirect path for security handler.');
                } else{
                    res.render('shared/' + status, { url: req.url, title: status + ' - ' + msg, message: msg }); // try default as last resort.
                }
            }            

        }

    };
     

};
