
'use strict';

var _ = require('lodash')
  , pkg = require('./package');

/**
* Updates, saves, modifies apps.json.
* @module appdate
*/
module.exports = {

    /**
    * Checks for a duplicate application.
    * @function
    * @param {object} apps - object of instaled apps.
    * @param {string} name - the app name to check for duplication.
    * @param {boolean} remove - optional arg to remove if exists.
    * @returns {object}
    */
    isDuplicate: function (apps, name, remove) {
        remove = remove || false;
        var exists = apps[name] ? true : false;
        if (exists && remove) {
            delete apps[name];
        }
        return { exists: exists, apps: apps };
    },

    /**
     * Creates an application object.
     * This saves key config info to apps.json. Makes it possible to start app from any dir.
     * TODO decide what to track whole config overkill but maybe other props need to be tracked.
     *
     * @function
     * @param {object} config - the loaded config object.
     * @param {boolean} removePaths - whether or not to trim uneeded paths for saving to apps.json
     * @returns {object}
     */
    get: function (config, removePaths) {

        removePaths = removePaths || false;

        var app = {};

        app.version = config.info.version;
        app.appName = config.appName;
        app.viewEngine = config.viewEngine;
        app.paths = config.paths;   
        app.angular = { enabled: config.angular.enabled, enabledBootstrapper: config.angular.enabledBootstrapper };
        app.less = { enabled: config.less.enabled }
        app.session = { key: config.session.key, secret: config.session.secret };
        app.log = config.log 
        app.linker = config.linker;
        app.fixups = config.fixups;
        app.lastLoaded = config.lastLoaded;

        delete app.linker['templates'];
        delete app.linker['files'];

        if (removePaths) app = this.deleteUnusedPaths(app);

        return app;
    },

    /**
     * Deletes unneeded paths in config that are populated in init.
     * @function
     * @param {object} app - the application object.
     * @returns {object}
     */
    deleteUnusedPaths: function (app) {
        if (!app.angular) delete app.paths['angular'];
        if (!app.less) delete app.paths['less'];
        delete app.paths['exec'];
        delete app.paths['root'];      
        delete app.paths['rootEscaped'];
        delete app.session['supported'];     
        return app;
    },

    /**
   * Saves the app back to apps.json.
   * @function
   * @param {object} config - the configuration object.
   * @param {boolean} saveOnly - indicates we are not updating.
   * @returns {object}
   */
    save: function (config, saveOnly) {

        var tmpConfg, app, root;
        saveOnly = saveOnly || false;
        root = config.root || config.paths.root;

        if (!saveOnly) {
            /* remove the app we'll add it back in a bit */
            delete config.apps[config.appName];

            /* need to create a clone of our config object. */
            tmpConfg = _.clone(config, true);
            app = this.get(tmpConfg, true);

            /* add the updated app back to the apps object */
            config.apps[app.appName] = app;
        }      

        /* save the apps object back to apps.json */
        pkg.saveSync(root + '/apps.json', config.apps);

        return config.apps;
    }

};
