﻿
'use strict';

var fs = require('fs')
  , fsextra = require('fs.extra')
  , path = require('path')
  , http = require('http')
  , url = require('url');


/**
* Disk IO operations module.
* @module diskio
*/
module.exports = {

    /**
    * Reads a file by specified source asynchronously.
    * @function read
    * @param {string} source - path to file.
    * @param {object} options - file options e.g. encoding: 'utf8'
    * @param {function} callback - callback function.
    */
    read: function read(source, options, callback) {
        options = options || 'utf8';
        fs.readFile(source, options, function (err, data) {
            if (err) throw err;
            else if (callback) {
                callback(data);
            }
        });
    },

    /**
    * Reads a file by specified source synchronously.
    * @function readSync
    * @param {string} source - path to file.
    * @param {object} options - file options e.g. encoding: 'utf8'
    * @returns {string} 
    */
    readSync : function readSync (source, options) {
        options = options || 'utf8'; 
        return fs.readFileSync(source, options);        
    },

    /**
    * Reads directory and returns all files.
    * @function readDir
    * @param {string} source - directory.
    * @param {function} callback - callback if files are found.
    */
    readDir: function readDir (source, callback) {
        fs.rreaddir(source, function (err, files) {
            if (err) throw err;
            if (callback) callback(files);
        });       
    },

    /**
   * Reads directory and returns all files synchronously
   * @function readDirSync
   * @param {string} source - directory.
   */
    readDirSync: function readDirSync (source) {
        return fs.rreaddir(source);
    },

    /**
    * Writes a file by specified source and data asynchronously.
    * @function write
    * @param {string} source - path to file.
    * @param {string|buffer} data - the data to write to file.
    * @param {function} callback - callback function.
    * @param {object} options - write file options.
    */
    write : function write(source, data, callback, options) {
        options = options || 'utf8';
        fs.writeFile(source, data, options, function (err) {
            if (err) throw err;
            else if (callback) callback(true, source);
        });
    },

    /**
    * Writes a file by specified source and data synchronously.
    * @function writeSync
    * @param {string} source - path to file.
    * @param {string|buffer} data - the data to write to file.
    * @param {object} options - write file options.
    */
    writeSync : function writeSync(source, data, options) {
        options = options || 'utf8';
        fs.writeFileSync(source, data, options);     
    },

    /**
    * Downloads a file from the web asynchronously.
    *
    * @function download
    * @param {string} source - path to file.
    * @param {string} target - the location to save the file.
    * @param {function} callback - callback function.
    * @param {object} options - sets flags, rename, encoding, reportProgress.
    */
    download: function download(source, target, callback, options ) {

        if (!source || !target) throw new Error('Invalid source or target specified.');

        var filename, file, opt, parsedUrl;
        options = options || {};
        reportProgress = options.reportProgress || true;
        parsedUrl = url.parse(source);
        filename = options.rename || parsedUrl.pathname.split('/').pop();  

        this.exists(target, function (exists) {
            if (!exists) {
                this.mkdir(target, function () {
                    httpGet();
                });
            } else {
                httpGet();
            }          
        });

        file = fs.createWriteStream(target + '/' + filename, options.flags || {});
        function httpGet() {
            var progress = { length: 0, completed: 0, remaining: 0, finishes: false, src: source, filename: filename };
            http.get(source, function (res) {
                if (options.encoding)
                    res.setEncoding(options.encoding);
                var headers = res.headers;
                progress.length = headers['content-length'] || 0;
                progress.remaining = progress.length;
                res.on('data', function (data) {                   
                    file.write(data);
                    progress.remaining -= data.length;
                    progress.completed += data.length;
                    if (reportProgress) {
                        process.stdout.clearLine();
                        process.stdout.cursorTo(0);
                        process.stdout.write('Downloading "' + filename + '" completed ' +
                            (100.0 * progress.completed / progress.length).toFixed(2) + '% ' +
                            (progress.length > 1000 ? (progress.length / 1000).toFixed(0) : 1) + ' kb');
                    }                      
                }).on('end', function () {
                    file.end();
                    progress.finished = true;
                    if (callback) callback(progress);
                }).on('error', function (err) {
                    if (err) throw err;
                });                
            });                 
        };        
    },

    /**
    * Makes a new directory asynchronously.
    * @function mkdir
    * @param {string} source - path to file.
    * @param {function} callback - callback function.
    */
    mkdir : function mkdir (source, callback) {
        fsextra.mkdirp(source, function (err) {
            if (err) throw err;
            else if (callback) callback();
        });    
    },

    /**
    * Makes a new directory synchronously.
    * @function mkdirSync
    * @param {string} source - path to file.
    */
    mkdirSync: function mkdirSync(source) {
        fsextra.mkdirpSync(source);   
    },

    /**
    * Copies source directory to target directory asynchronously.
    * @function copy
    * @param {string} source - path to source file.
    * @param {string} target - path to target file.
    * @param {function} callback - callback function.
    */
    copy : function copy (source, target, callback) {
        fsextra.copy(source, target, function (err) {
            if (err) throw err;
            else if (callback) callback();
        });
    },

    /**
    * Copies source directory to target directory recursively and asynchronously.
    * @function copyRecursive
    * @param {string} source - path to source file.
    * @param {string} target - path to target file.
    * @param {function} callback - callback function.
    */
    copyRecursive: function copyRecursive(source, target, callback) {
        fsextra.copyRecursive(source, target, function (err) {
            if (err) throw err;
            else if (callback) callback(source, target);
        });        
    },

    /**
    * Move source directory to target directory asynchronously.
    * @function move
    * @param {string} source - path to source file.
    * @param {string} target - path to target file.
    * @param {function} callback - callback function.
    */
    move : function move(source, target, callback) {
        fsextra.move(source, target, function (err) {
            if (err) throw err;
            else if (callback) callback(true, source, target);
        });
    },

    /**
    * Remove directory asynchronously.
    * @function remove
    * @param {string} source - path to source file.
    * @param {function} callback - callback function.
    */
    remove : function remove (source, callback) {
        fsextra.rmrf(source, function (err) {
            if (err) throw err;
            else if (callback) callback(true, source);
        });    
    },

    /**
    * Remove directory synchronously.
    * @function removeSync
    * @param {string} source - path to source file.
    */
    removeSync : function removeSync (source) { 
       fsextra.rmrfSync(source);        
    },

    /**
     * Walk directories and remove files/dirs by criteria. Specify critera by 
     *
     * TODO make more compatible with other scenarios.
     *
     * @function removeRecursiveFilter
     * @param {string} source - source to start walking from.
     * @param {array} criteria - the criteria by which to find and remove by.
     * @param {function} callback - callback when finished.
     * @param {string} options - RegEx options. e.g. 'gi' for global ignore case.
     */
    removeRecursiveFilter: function removeRecursiveFilter(source, filters, callback, options) {

        var self, w, removed, count;
        self = this;
        options = options || 'gi';
        removed = [];
        count = 0;

        if (!this.existsSync(source) || !filters || !callback) throw new Error('Filter and callback are required and path must exist.');
        
        w = fsextra.walk(source);
        w.on('file', action);
        w.on('directories', action);
        w.on('errors', function (root, stats, next) {
            console.log(' > Error walking ' + root);
            next();
        });
        w.on('end', function () {
            callback(removed, count);
        });

        function action(root, stats, next) {          
            for (var i = 0; i < filters.length; i++) {
                var search, regEx;
                search = filters[i];
                regEx = new RegExp(search, options);
                if (regEx.test(stats.name) || regEx.test(root)) remove();
                if (i == (filters.length - 1)) next();
            }
            function remove() {
                var curPath = stats.type == 'directory' ? root : root + '/' + stats.name;              
                self.remove(curPath);                   
                removed.push(curPath);
                count += 1;        
            }
        }
    },

    /**
    * Checks if directory exists asynchronously.
    * @function exists
    * @param {string} source - path to source file.
    * @param {function} callback - callback function.
    */
    exists : function exists(source, callback) {
        fs.exists(source, function (exists) {
            if (callback) callback(exists);
        });    
    },

    /**
    * Checks if directory exists synchronously.
    * @function existsSync
    * @param {string} source - source to source file.
    * @returns {boolean}
    */
    existsSync : function existsSync (source) {
        return fs.existsSync(source);
    },

    /**
    * Watch directory asynchronously.
    * @function watch
    * @param {string} source - path to source file.
    * @param {function} callback - callback function.
    */
    watch : function watch(source, callback) {
        fs.watch(source, function (event, filename) {
            if (callback && filename) callback(event, filename);
            else throw new Error('Unable to watch source "' + source + '"');
        });
    },

    /**
    * Resolves path synchronously.
    *
    * TODO not sure why the heck this is here figure out why.
    *
    * @function resolve
    * @param {string} source - source path to resolve.
    * @returns {string} 
    */
    resolve : function resolve (source) {
        return path.resolve(source);
    }


};




