
'use strict';

var self = this
  , ejs = require('ejs')
  , diskio = require('./diskio')
  , include = require('include-all')
  , pkg = require('./package')
  , _ = require('lodash');
  //, ProgressBar = require('progress');

/**
* Miscellaneous helpers module.
* @module helpers
*/
var helpers = module.exports = {

    /**
   * Get JavaScript type.
   * @function getType
   * @param {string|object|array|boolean|number|undefined|null|regex} obj - description. 
   * @returns {boolean} 
   */
    getType: function getType(obj) {
        return Object.prototype.toString.call(obj).replace(/^\[object (.+)\]$/, "$1").toLowerCase();
    },

    /**
    * Checks if obj is JSON.
    * @function isJSON
    * @param {string|object} obj - object to check for JSON.
    * @returns {boolean} 
    */
    isJSON: function isJSON(obj) {
        try {
            JSON.parse(obj);
        } catch (e) {
            return false;
        }
        return true;
    },

    /**
    * Reads a file then uses ejs with data object to render.
    * @function render
    * @param {string} source - seed file path.
    * @param {object} obj - data object used to render seed file.
    * @returns {string|null}
    */
    render: function render (source, obj) {

        var data, file;
        obj = obj || {};
        obj.filename = source;

        /* make sure the file exists */
        if (!diskio.existsSync(source)) throw new Error(' > Unable to locate seed file at "' + source + '"');

        /* render the file using ejs and passing obj */
        data = diskio.readSync(source);    
          
        file = ejs.render(data, obj );
        return file;

    },

    /**
     * Removes an object from an Array collection by property name and value.
     * @function removeByProperty
     * @param {array} collection - the collection to search.
     * @param {string} prop - the property to find.
     * @param {string} value - the property value to match. 
     * @returns {array}
     */
    removeByProperty: function removeByProperty(collection, prop, value) {
        for (var i = 0; i < collection.length; i++) {
            if (collection[i][prop] == value) collection.splice(i, 1);            
        }
        return collection;
    },

    /**
   * Removes all objects by array of types or single type as string.
   * @function removeLinkerByType
   * @param {array} collection - the collection of objects.
   * @param {array|string} types - the property to find.
   * @returns {array}
   */
    removeLinkerByType: function removeLinkerByType(collection, types) {
        var result;
        if(_.isArray(types)) {
            result = _.filter(collection, function (obj, idx) {
                var valid = true;
                for (var i = 0; i < types.length; i ++) {
                    if (obj.type === types[i]) valid = false;
                }
                return valid;
            });            
        } else {
           result = _.filter(collection, function (obj, idx) {
                return obj.type !== types;
            });
        }
        return result;
    },

    /**
    * Gets an object by property namespace using dot notation ex: "user.name.first".
    *
    * @function getPropertyByName
    * @param {array} obj - the object to search.
    * @param {string} prop - the property to find.
    * @param {string|array} strip - array of replacement expression to be matched at each dot notation or a single string.
    * @param {boolean} matchCase - whether to match casing of property name.    
    * @param {string} def - the default value to return if not found.   
    * @returns {object|null|string|boolean|number}
    */
    getPropertyByName: function getPropertyByName(obj, prop, strip, caseSensitive, def) {

        def = def === undefined ? null : def;
        caseSensitive = caseSensitive || false;
        strip = strip || false;

        for (var i = 0, prop = prop.split('.'), len = prop.length; i < len; i++) {

            if (!obj || typeof obj !== 'object') return def;

            obj = _.find(obj, function (property, key) {
                if (!strip && caseSensitive) return key === prop[i];
                if (!strip) {
                    return key.toLowerCase() === prop[i].toLowerCase();
                } else {
                    strip = _.isArray(strip) ? strip[i] : strip;
                    var tmpKey = key.replace(strip, '');        
                    return tmpKey.toLowerCase() === prop[i].toLowerCase();
                }                    
            });
                                    
        }

        if (obj === undefined) return def;

        return obj;
    },

    ///**
    //* Starts a progress bar for a process.
    //* incomplete and complete templates should match in length.
    //* @function showProgress
    //* @param {number} ticks - number of ticks for progress bar.
    //* @param {string} completeMsg - the completion message.
    //* @param {number} width - max width of the progress bar.
    //* @param {boolean} initialSpace - add space to initial tick.
    //* @returns {object} 
    //*/
    //showProgress: function showProgress(ticks, completeMsg, width) {

    //    /* if ticks is string assume its complete message */
    //    if (this.getType(ticks) == 'string') {           
    //        completeMsg = ticks;
    //        ticks = null;
    //    }

    //    var bar, timer, complete, completeStart;
    //    ticks = ticks || 150;
    //    completeMsg = completeMsg || '';
    //    complete = '>';
    //    width = width || 30;

    //    if (complete.length > 1) width -= complete.length;

    //    bar = new ProgressBar(':bar', { total: ticks, width: width, complete: complete, incomplete: ' ', prefix: ' ' });
    //    timer = setInterval(function () {      
    //        bar.tick();
    //        if (bar.complete) {
    //            if(completeMsg)
    //                console.log(completeMsg);
    //            clearInterval(timer);
    //        }
    //      }, 100);
    //    return { interval: timer, msg: completeMsg };
    //},

    ///**
    // * Stops a progress bar for a process.
    // * @function stopProgress
    // * @param {object} progress - the progress bar interval/msg options object.
    // */
    //stopProgress: function stopProgress(progress) {
    //    if (progress.interval) {
    //        clearInterval(progress.interval);
    //        if(progress.msg)
    //            console.log(progress.msg);
    //    }         
    //},

    /**
   * Converts string to proper casing.
   * @function stringToCase
   * @param {string} str - the string to convert.
   * @param {string} type - the type to convert to, e.g. title, first, camel, pascal.
   * @returns {string}
   */
    stringToCase: function stringToCase(str, type) {

        if(type == 'title')
            return str.replace(/\w\S*/g, function (txt) { return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase(); });

        if(type == 'first')
            return str.charAt(0).toUpperCase() + str.slice(1);

        if (type == 'camel') {
            return str.toLowerCase().replace(/-(.)/g, function (match, group1) {
                return group1.toUpperCase();
            });
        }

        if (type == 'pascal') {
            return str.replace(/\w+/g, function (w) { return w[0].toUpperCase() + w.slice(1).toLowerCase(); });
        }

    },

    /**
    * Truncates string if max length is exceeded.
    * @function stringTruncate
    * @param {string} str - the string to truncate.
    * @param {number} type - the type to convert to, e.g. title, first, camel, pascal.
    * @param {string} str - the trailing string when max is exceeded.
    * @returns {string}
    */
    stringTruncate: function stringTruncate(str, max, trailing) {
        if (!max) return str;
        var result = ''
          , trailLen = !trailing ? 0 : trailing.length
          , max = max - trailLen;
        if (str.length > max) {
            str = str.slice(0, max);
            str += (trailing || '');
            return str;
        }
        return str;
    },

   /**
  * Gets the string value of a property key.
  * @function propertyNameToString
  * @param {object} obj - the object of properties.
  * @param {string} prop - the property name to lookup.
  * @param {string} def - the default value if property is not found.
  * @returns {string}
  */
    propertyNameToString: function propertyNameToString(obj, prop, def) {
        def = def || null;
        for (var key in obj) {
            if (!obj.hasOwnProperty(key)) continue;
            if (key == prop)
                return obj[key];
        }
        return def;
    },

    /**
    * Looks up a handle by name in an object. If the property name is a function it returns it.
    * @function lookupHandle
    * @param {object} obj - the object of properties.
    * @param {string} prop - the property name to lookup.
    * @param {string} def - the default value if property is not found.
    * @returns {function}
    */
    lookupHandle: function lookupHandle(obj, prop, def) {

        /* nothng to do just return it back */
        if (_.isFunction(prop)) return prop;

        ///* callback name specified look it up */
        //var handle = _.find(obj, function (handler, key) {
        //    this.getPropertyByString()
        //    return key.toLowerCase() === prop.toLowerCase();
        //});

        var handle = this.getPropertyByName(obj, prop);

        if (!handle) return def;

        return handle;

    },
    
    /**
    * Simple function to sort and object.
    * @function sortObjectByKey
    * @param {object} obj - the object of properties.
    * @param {string} type - the type to sort by. "keys or values".
    * @returns {object}
    */
    sortObject: function sortObject(obj, type) {
        if (!obj) return;
        type = type || 'keys';
        var tmp = {},
            vals = (type === 'keys' ? _.keys(obj) : _.values(obj));
        vals.sort();
        for (var i = 0; i < vals.length; i++) {
            var k;
            if (type === 'keys') {
                k = vals[i];
                tmp[k] = obj[k];
            } else {
                k = Object.keys(obj).filter(function(key) {
                     return obj[key] === vals[i];
                })[0];
                tmp[k] = vals[i];
            }
        }
        obj = tmp;
        return obj;
    },
    
    /**
   * Advanced sorting function including use of primers.
   * Properties in form of [name, [age, 'desc']]
   * @function sortArray
   * @param {array} objArray - the array object.
   * @param {array} properties - array of properties to sort by.
   * @returns {array}
   */
    sortArray: function (objArray, properties /*, primers*/) {

        var primers = arguments[2] || {}; // primers are optional

        properties = properties.map(function (prop) {
            if (!(prop instanceof Array)) {
                prop = [prop, 'asc'];
            }
            if (prop[1].toLowerCase() == 'desc') {
                prop[1] = -1;
            } else {
                prop[1] = 1;
            }
            return prop;
        });

        function valueCmp(x, y) {
            return x > y ? 1 : x < y ? -1 : 0;
        }

        function arrayCmp(a, b) {
            var arr1 = [], arr2 = [];
            properties.forEach(function (prop) {
                var aValue = a[prop[0]],
                    bValue = b[prop[0]];
                if (typeof primers[prop[0]] != 'undefined') {
                    aValue = primers[prop[0]](aValue);
                    bValue = primers[prop[0]](bValue);
                }
                arr1.push(prop[1] * valueCmp(aValue, bValue));
                arr2.push(prop[1] * valueCmp(bValue, aValue));
            });
            return arr1 < arr2 ? -1 : 1;
        }

        objArray.sort(function (a, b) {
            return arrayCmp(a, b);
        });

        return objArray;
    },

    /**
    * Takes in a juggling db model and strips hidden properties.
    * @function trimModel
    * @param {object} model - array of properties to sort by.
    * @param {string|object} xModel - the xModel the model belongs to or its string name to be looked up.
    * @returns {array|object}
    */
    trimModel: function trimModel (model, xModel) {

        /* if xModel is string look it up */
        if (_.isString(xModel)) {
            if (!_xmodels) return model;
            xModel = xModel.replace(/Model/i, ''); 
            xModel = this.getPropertyByName(_xmodels, xModel, /Model/i, false, {});
        }   

        /* make sure we have a plain object */
        model = JSON.parse(JSON.stringify(model));
  
        if (_.isArray(model)) {
            var tmpArr = [];
            _.forEach(model, function (m) {
                var tmp = _.forEach(m, function (value, key) {
                    if (xModel && xModel.properties[key] && xModel.properties[key].hidden === true)
                        delete m[key];
                });
                tmpArr.push(tmp);
            });
            model = tmpArr;
        } else {
            model = _.forEach(model, function (value, key) {             
                if (xModel && xModel.properties[key] && xModel.properties[key].hidden === true)
                    delete model[key];
            });
        }
        return model;


    },

    existsByObject: function existsByObject(collection, obj, def) {
        def = def || false;
        var exists = _.find(collection, obj);
        return exists || def;
    },    

    /* we don't need all helpers exposed so we expose a few public in globals */
    'public': {

        trimModel: function trimModel(model, xModel) {
            return helpers.trimModel(model, xModel);
        },

        getPropertyByName: function (obj, keys, strip, caseSensitive, def) {
            return helpers.getPropertyByName(obj, keys, strip, caseSensitive, def);
        },

        sortArray: function sortArray(objArray, properties, primers) {        
            return helpers.sortArray(objArray, properties, primers);
        },

        sortObject: function sortObject(obj, type) {
            return helpers.sortObject(obj, type);
        },

        propertyNameToString: function propertyNameToString(obj, prop, def) {
            return helpers.propertyNameToString(obj, prop, def);
        },

        stringTruncate: function stringTruncate(str, max, trailing) {
            return helpers.stringTruncate(str, max, trailing);
        },

        stringToCase: function stringToCase(str, type) {
            return helpers.stringToCase(str, type);
        }

    }

};
