
'use strict';

var diskio = require('./diskio')
  , _ = require('lodash');

/**
* Reads/saves package.json configuration files.
* @module pkg
*/
module.exports = {

    /**
    * Gets a configuration file by path.
    * @function
    * @param {string} source - path to the package file.
    * @returns {string}
    */
    get: function (source) {     
        var src = diskio.readSync(source);
        return JSON.parse(src.toString());
    },

    /**
    * Gets a configuration property by path.
    * @function
    * @param {string} source - the path to the package file.
    * @param {string} prop - the property name to retrive.
    * @returns {string}
    */
    getProperty: function (source, prop) {
        var data = this.get(source);
        return data[prop] || 'undefined';
    },

     /**
    * Saves a configuration file by path and data asynchronously.
    * @function
    * @param {source} - source to the configuration file.
    * @param {string|buffer} data - the data to write to file.
    * @param {function} callback - callback when finished.
    * @param {object} options - options for wirting file to disk.
    */
    save: function (source, data, callback, options) {
        if (!data || !callback) return;
        /* stringify with pretty formatting. */   
        data = JSON.stringify(data, null, 4);
        diskio.write(source, data, callback, options);
    },

    /**
    * Saves a configuration file by path and data synchronously.
    * @function
    * @param {source} - source to the configuration file.
    * @param {string|buffer} data - the data to write to file.
    * @param {object} options - options for wirting file to disk.
    */
    saveSync: function (source, data, options) {
        if (!data) return;
        /* stringify with pretty formatting. */
        data = JSON.stringify(data, null, 4);
        diskio.writeSync(source, data, options);
    },

    /**
   * Creates a package file by specified source path and config object synchronously.
   * @function
   * @param {string} - source to the configuration file.
   * @param {object} data - the data to write to file.
   * @param {object} options - options for wirting file to disk.
   * @returns {boolean} 
   */
    createSync: function (source, data, options) {
        if (!data) return;
        var defaults = {
            name: 'App'
          , 'private': true
          , version: '0.0.1'       
          , description: 'Xipframe MVC framework application.'
          , author: {
              name: 'Your name'
            , email: 'youremail@domain.com'
          }
          , dependencies: {
              xipframe: '~0.2.4',             
              grunt: '~0.4.1'
          }
          , scripts: {
              start: 'node server.js',
              debug: 'node debug server.js'
          }     
          , readmeFilename: "README.md"
          , keywords: [
              'Xipframe'
            , 'MVC'
            , 'MVC Framework'
            , 'REST'
            , 'API'
            , 'Node'
            , 'Express'
          ],
          main: 'server.js'
        };
        
        /* Merge in the default and passed config */   
        _.merge(defaults, data);      

        this.saveSync(source, defaults, options);
    },



};
