
'use strict';

/**
 * Error stacktrace helper.
 * @module stacktrace
 * @returns {object}
*/
module.exports = {

        /**
        * Returns formatted error and stacktrace.
        * @funciton stackFormat      
        * @param {object} exception - the error exception to format.
        * @param {boolean} includeStackStr - if true adds a stackStr property where stack is convert to readable string.
        * @param {string} brk - the line break to use when using includingStackStr.
        * @returns {object}
        */
        format: function (exception, includeStackStr, brk) {

      
            var err, stack, file;
            err = {};
            err.name = exception.name || 'Unknown Exception';
            err.message = exception.message || 'Server encountered fatal exception error!';
            err.metadata = {};
            err.codes = exception.codes || null;

            stack = this.stackParse(exception);
            err.metadata.stack = stack || {};

            if (stack && stack[0]) {

                err.metadata.function = stack[0].getFunctionName() || 'Unknown function';
                err.metadata.line = stack[0].getLineNumber() || 0;
                err.metadata.col = stack[0].getColumnNumber() || 0;
                file = stack[0].getFileName() || 'Unknown filename';

                /* remove some unecessary chars need to make this better i know */
                file = file.replace('[as handle]', '');
                file = file.replace(' ', '');
                file = file.replace(/(\(|\))/, '');
                err.metadata.file = file;

                if (includeStackStr)
                    err.metadata.stackString = this.stackToString(stack, brk);

            }
          

            return err;
                  

        },

        /**
        * Formats stack trace object to string
        * @funciton stackToString 
        * @param {object} stack - the stack to format to string.
        * @param {string} brk - the line break to use, usually '\n' or '<br/>' defaults to '\n'.
        * @returns {string}
        */
        stackToString: function(stack, brk){
            var str = '';
            brk = brk || '\n';
            stack.forEach(function (frame) {
                str += ('at ' + (frame.functionName || '') + ' ' + frame.fileName + ':' + (frame.lineNumber || 0) + ':' + (frame.columnNumber || 0) + brk);
            });
            return str;
        },

          /**
         * Gets V8 stacktrace
         * @funciton stackGet 
         * @returns {object}
         */
        stackGet: function (belowFn) {
            var oldLimit = Error.stackTraceLimit;
            Error.stackTraceLimit = Infinity;

            var dummyObject = {};
            Error.captureStackTrace(dummyObject, belowFn || exports.get);

            var v8Handler = Error.prepareStackTrace;
            Error.prepareStackTrace = function (dummyObject, v8StackTrace) {
                return v8StackTrace;
            };

            var v8StackTrace = dummyObject.stack;
            Error.prepareStackTrace = v8Handler;
            Error.stackTraceLimit = oldLimit;

            return v8StackTrace;
        },

        /**
        * Parses V8 stacktrace using callsite
        * @funciton stackParse      
        * @returns {object}
        */
        stackParse: function (err) {
            if (!err.stack) {
                return [];
            }

            var self = this;
            var lines = err.stack.split('\n').slice(1);

            return lines
              .map(function (line) {
                  if (line.match(/^\s*[-]{4,}$/)) {
                      return self._createParsedCallSite({
                          fileName: line,
                          lineNumber: null,
                          functionName: null,
                          typeName: null,
                          methodName: null,
                          columnNumber: null,
                          'native': null,
                      });
                  }

                  var lineMatch = line.match(/at (?:([^\s]+)\s+)?\(?(?:(.+?):(\d+):(\d+)|([^)]+))\)?/);
                  if (!lineMatch) {
                      return;
                  }

                  var object = null;
                  var method = null;
                  var functionName = null;
                  var typeName = null;
                  var methodName = null;
                  var isNative = (lineMatch[5] === 'native');

                  if (lineMatch[1]) {
                      var methodMatch = lineMatch[1].match(/([^\.]+)(?:\.(.+))?/);
                      object = methodMatch[1];
                      method = methodMatch[2];
                      functionName = lineMatch[1];
                      typeName = 'Object';
                  }

                  if (method) {
                      typeName = object;
                      methodName = method;
                  }

                  if (method === '<anonymous>') {
                      methodName = null;
                      functionName = '';
                  }

                  var properties = {
                      fileName: lineMatch[2] || null,
                      lineNumber: parseInt(lineMatch[3], 10) || null,
                      functionName: functionName,
                      typeName: typeName,
                      methodName: methodName,
                      columnNumber: parseInt(lineMatch[4], 10) || null,
                      'native': isNative,
                  };

                  return self.stackParseCallSite(properties);
              })
              .filter(function (callSite) {
                  return !!callSite;
              });
        },
        
        /**
        * Parses V8 callsite
        * @funciton stackParseCallSite       
        * @returns {object}
        */
        stackParseCallSite: function (properties) {
            var methods = {};
            for (var property in properties) {
                var prefix = 'get';
                if (property === 'native') {
                    prefix = 'is';
                }
                var method = prefix + property.substr(0, 1).toUpperCase() + property.substr(1);

                (function (property) {
                    methods[method] = function() {
                        return properties[property];
                    };
                })(property);
            }

            var callSite = Object.create(methods);
            for (var property in properties) {
                callSite[property] = properties[property];
            }

            return callSite;
        }

};

   

