 var express = require('express')
  , path = require('path')
  , http = require('http');

/**
* Configures Express.
* @module express
* @param {object} xipframe - the loaded configuration.
*/
module.exports = function (xipframe) {

    'use strict';

    var app = express()
      , config = xipframe.config
      , handlers = xipframe.handlers
      , userHandlers = xipframe.userHandlers
      , routes = xipframe.routes
      , _ = xipframe.utils._
      , userErrorHandler = userHandlers.error ? true : false
      , userFnfHandler = userHandlers.fnf ? true : false
      , InitError = xipframe.errors.InitError
      , store;

    app.configure(function () {

        /* set host/port to start on */
        app.use(express.favicon(config.paths.public + '/img/favicon.ico'));
        app.set('host', process.env.HOST || config.host);
        app.set('port', process.env.PORT || config.port);

        /* configure view setup */
        app.set('views', config.paths.views);

        /* set ejs-locals for layout support */
        if (config.viewEngine === 'ejs' || config.viewEngine === 'html') {          
            app.engine(config.viewEngine, require('ejs-locals'));
            app.set('view engine', 'ejs');
        }          

        /* using Angular likely and want to use plain html ext. */
        //if (config.viewEngine == 'html')
           // app.engine('html', require('ejs').renderFile);

        /* set the view engine type */
        app.set('view engine', config.viewEngine);

        /* define cookie parser if enabled */
        if (config.express.cookieParser) app.use(config.express.cookieParser);

        // Redo the session handling too verbose as supported adapters expands/hacky need to fix up, only need the time!

        /* check for supported session adapter. */
        if (!_.contains(config.session.supported, config.session.adapter)) throw new InitError(config.session.adapter + ' not found or is not supported.');

        /* check for memory session adapter */
        if (config.session.adapter == 'memory') {
            var MemoryStore = express.session.MemoryStore;
            store = new MemoryStore();
        }

        /* check if mongodb is session adapter */
        if (config.session.adapter == 'mongodb') {
       
            var MongoStore = require('connect-mongo')(express)
                , settings = config.session[config.session.adapter] || config.adapters[config.session.adapter] || null;
            if (!settings) throw new InitError('Session adapter settings for' + config.session.adapter + ' could not be found.');   
            settings.db = settings.database;
            store = new MongoStore(settings);
        }

        /* check for redis sessions */
        if (config.session.adapter == 'redis') {
            var RedisStore = require('connect-redis')(express)
              , settings = config.session[config.session.adapter] || config.adapters[config.session.adapter] || null;
            if (!settings) throw new InitError('Session adapter settings for' + config.session.adapter + ' could not be found.');
            settings.db = settings.database;
            settings.pass = settings.password || null;
            store = new RedisStore(settings);
        }

        if (!store) throw new InitError('Sessions could not be initialized, xipframe requires sessions and cannot start...');

        /* setup session store */
        xipframe.sessions = {
            secret: config.session.secret,
            key: config.session.key,
            cookie: { httpOnly: config.session.httpOnly, maxAge: config.session.maxAge },
            store: store
        };

        /* set the express sessions */
        app.use(express.session(xipframe.sessions));

        /* define body parser if enabled */
        if (config.express.bodyParser) app.use(config.express.bodyParser);

        /* define RESTful methods if enabled */
        if (config.express.methodOverride) app.use(config.express.methodOverride);    
               
        /* locals handler add before middleware for overriding */
        app.use(handlers.locals(xipframe));

        /* security handler */
        app.use(handlers.security(xipframe));

        /* enables redirects via routes if enabled */
        app.use(handlers.redirects(xipframe));

        /* define csrf protection if enabled */
        if (config.express.csrf) {
            app.use(config.express.csrf);
            app.use(handlers.csrf(xipframe));
        }

        /* middlware */
        _.forEach(xipframe.middleware, function (m) {
            if (_.isFunction(m)) app.use(m);
        });

        /* router */
        app.use(app.router);

        /* serve statics */
        app.use(express.static(config.paths.public));
        
        /* add static for angular if enabled */
        if (config.angular.enabled) {
            app.use(express.static(config.paths.angularPublic));
            app.use(express.static(config.paths.views));
        }

        /* error handler */
        app.use(handlers.error(xipframe));

        /* file not found handler */
        app.use(handlers.fnf(xipframe));

        /* user config handles errors */
        if (userErrorHandler && !config.handlers.error.enabled) app.use(userHandlers.error);

        /* user config handles 404 */
        if (userFnfHandler && !config.handlers.fnf.enabled) app.use(userHandlers.fnf);

    });
    /* add user defined and auto routes to express */
    _.forEach(routes, function (route) {  
        var normalizePath = route.path.split(' ')[1];     
        app[route.method](normalizePath, route.action);
    });

    /* include user defined express config */
    if (config.express.extend) config.express.extend(app, express);

    /* set xipframe express instance */
    xipframe.express = app;

};
