

var Xipframe = (function () {

    'use strict';
    
    var cluster = require('./cluster') // NOT IMPLEMENTED - cluster unstable
      , path = require('path')
      , utils = require('./utils')
      , build = require('./build')
      , configuration = require('./configuration')
      , webserver = require('./webserver')
      , handlers = require('./handlers')
      , IO = require('./io')
      , cli = require('./cli')
      , _ = utils._
      , _global = global;

    
    /**
    * Xipframe engine class.
    * @class Xipframe
    * @param {object} config - user defined config is optional.
    */
    function Xipframe() {

        var initConfig, rootPath, self, commands;
        self = this;

        /* need the root of the lib */
        root = path.join(__dirname, '../');

        if (!utils.diskio.existsSync(root + '/apps.json'))
            utils.diskio.writeSync(root + '/apps.json', JSON.stringify({}));
        
        /* get the needed initialization paths/details */
        initConfig = {
            
            apps: utils.package.get(root + '/apps.json') || {},
            info: utils.package.get(root + '/package.json'),
            basename: path.basename(process.cwd()).toLowerCase(),
            platform: require('os').platform(),
            pid: process.pid,
            paths: {
                root: root,
                exec: process.cwd()
            }
            
        };

        commands = {
            install: 0,
            'new': 0,
            create: 0,
            generate: 1,
            gen: 1,
            start: 2,
            run: 2,
            uninstall: 3,
            'delete': 3,
            remove: 4,
            download: 5,
            diag: 6,
            diagnostics: 6,
            info: 7,
            'package': 7,
            help: 8,
            usage: 8
        };
    
        this.remote = false;
        this.config = {};
        this.log = {};  

        this.middleware = {};
        this.controllers = {};
        this.models = {};
        this.schema = {};
        this.routes = {};
        this.security = {};

        this.express = {};
        this.io = {};
        this.sessions = {};

        this.children = [];
        this.handlers = {};
        this.userHandlers = {};

        this.exiting = false;
        this.commands = commands;
        this.command = null;
        this.utils = utils;
        this.errors = configuration.errors;
     
        /* load in some basic paths */
        _.merge(this.config, initConfig);

        /* catch uncaught errors */
        process.on('uncaughtException', function (err) {
            err = utils.trace.format(err, true, '|');
            var name = err.name || 'Unknown Exception'
              , logMsg = '';

            console.log('');
            logMsg = ('[' + name + '] ' + err.message) +
                     ' line: ' + (err.metadata.line || 0) +
                     ' col: ' + (err.metadata.col || 0) +
                     '\nfile: ' + (err.metadata.file || 'unknown');            
            
            if (self.log && self.log.console) 
                self.log.console.error(logMsg)
            else
                console.log(logMsg);

            self.rollback();

        });

    };

    /**
    * Runs a Xipframe application by parsing the supplied arguments.|
    * @function run
    * @augments xipframe
    */
    Xipframe.prototype.run = function run() {
        cli.parse(this);
    };

    /**
    * Loads a configuration and the if all goes well boots server.
    * @function load
    * @param {string} appName
    * @param {boolean} force - if true forces upgrade.
    * @augments xipframe
    */
    Xipframe.prototype.load = function load(appName, force) {
        cli.load(appName, this, force);
        this.configure();
    };

    /**
    * Applies the loaded configuration to xipframe.
    * @function cofigure
    * @augments xipframe
    */
    Xipframe.prototype.configure = function configure() {

        /* get instance of Logger */
        this.log = utils.logger.get(this.config.log);

        console.log('');
        console.log(' > Xipframe is booting.');

        /* load framework modules. */
        console.log(' > Loading application framework.');

        /* include all other modules */
        var include = configuration.include(this);
        this.middleware = include.middleware();
        this.models = include.models();
        this.controllers = include.controllers();
        this.handlers = handlers;
        this.userHandlers = include.handlers();
        this.security = include.security();
        this.routes = include.routes();

        /* make sure xipframe/node have same env */
        process.env['NODE_ENV'] = this.config.environment;

        /* grunt and groan out some files */
        console.log(' > Configuring Grunt and linking assets.');
        this.grunt(this.config.environment);

        /* config is fully loaded start the server */
        this.start();
    };

    /**
    * Start Xipframe server.
    * @function start
    * @augments xipframe
    */
    Xipframe.prototype.start = function start () {

        var self, server, ssl, config;
        self = this;
        ssl = this.config.ssl;
        config = this.config;

        /* build framework routes, models etc. */
        console.log(' > Applying framework configuration.');
        build.framework(this);

        /* configure an express instance */
        console.log(' > Configuring webserver server.');
        webserver.express(this);

        /* if ssl make sure we have all we need*/
        if (ssl.enabled && ssl.key && ssl.cert) {
            var options = {}
              , https = require('https');
            options.key = utils.diskio.readSync(ssl.key);
            options.cert = utils.diskio.readSync(ssl.cert);
            server = https.createServer(options, this.express);
        } else {
            var http = require('http');
            server = http.createServer(this.express);
        }

        /* set the server instance*/
        this.express.server = server;
        this.express.connections = [];

        /* check for socket.io enabled */
        this.startSocket();

        /* set server instance once listening */
        server.on('listening', function () {
            
            /* listen for shutdown termination events. */
            webserver.shutdown(self)();
            
        });

        /*
        * on connection push to collection listen on close to remove.
        * if we shut down the server we'll iterate the connections if
        * any and destroy them to shutdown more gracefully.
        */
        server.on('connection', function (connection) {
            connection.on('close', function () {
                self.express.connections.splice(self.express.connections.indexOf(connection), 1);
            });
            self.express.connections.push(connection);

        });

         /* add enabled globals */
        build.globals(this)(_global);

        /* add models manually to maintain context */
        //_global['_models'] = this.schema.models;

        /* start the server listening */
        server.listen(config.port, config.host, function () {

            console.log('');
            console.log(' ###############################################################');
            console.log('  Xipframe Started');
            console.log(' ###############################################################');
            console.log('');
            console.log(' > Application "' + config.appName + '" powered by Xipframe running.');
            console.log(' > Database Adapter: "' + config.adapters.active + '"');
            console.log(' > Host: ' + config.host + '  Port: ' + config.port + ' Environment: ' + config.environment);
            console.log(' > View your application at --> http://' + config.host + ':' + config.port);

        });

    };


    /**
    * Starts socket.io
    * @function startSocket   
    * @augments xipframe
    */
    Xipframe.prototype.startSocket = function startSocket () {

        console.log(' > Starting socket.io server.');

        /* fire up socket.io if enabled */
        if (!this.config.io.enabled) return;

        /* start socket io */
        IO.start(this);

    };

    /**
    * Shuts down the Xipframe engine.
    * @function shutdown
    * @param {function} exit - analgous to the process.exit() function.
    * @augments xipframe
    */
    Xipframe.prototype.shutdown = function shutdown (exit) {

        var self = this;

        /* exit xipframe */
        this.exiting = true;

        /* if there's no server exit */
        if (!this.express.server) {
            if (exit) exit();
            return false;
        }

        try {
            
            console.log('');
            console.log(' ###############################################################');
            console.log('  Xipframe Shutting Down');
            console.log(' ###############################################################');

            /* close the server and then exit */
            console.log('');
            console.log(' > Server will close resources...please wait.');
            console.log('');

            /* kill any child processes */
            console.log(' > Closing ' + this.children.length + ' child processes.');
            if (this.children && this.children.length > 0) {
                _.forEach(self.children, function (child) {
                    process.kill(child.pid);
                });
            }

        } catch (e) {
            
            /*
            * TODO: dig into this more.
            *
            * throws error here sometimes haven't dug into it much, 
            * may be platform specific in some cases it seems so 
            * for now just catch.
            */
            

        } finally {

            setTimeout(function () {
                
                var openConnections = self.express.connections.length;
                console.log(' > Destroying ' + openConnections + ' open connections.');
                _.forEach(self.express.connections, function (connection) {
                    connection.destroy();
                });

                self.express.server.unref();
                
                self.express.server.close(function () {
                    
                    console.log(' > Server shutdown successfully.');
                    console.log('');
                    if (exit) exit();

                });

            }, 1100);


        }       

    };

    /**
    * Installs a new Xipframe based application.
    * @function install
    * @param {object} userConfig - the user defined configuration to build the new application with.
    * @augments xipframe
    */
    Xipframe.prototype.install = function install (userConfig) {
        /* run the install script */
        cli.install(this, userConfig);
    };

    /**
    * Uninstalls a Xipframe based application.
    * @function uninstall
    * @param {string} appName - the application name to be uninstalled or removed.
    * @param {boolean} jsonOnly - if true only removes app config from apps.json.
    * @augments xipframe
    */
    Xipframe.prototype.uninstall = function uninstall (appName, jsonOnly) {
        if (jsonOnly) cli.remove(appName, this);
        else cli.uninstall(appName, this);
    };

    /**
    * Gets Xipframe package information and installed apps.
    * @function info
    * @param {boolean} suppressConsole - only returns object instead of writing to console.
    * @augments xipframe
    */
    Xipframe.prototype.info = function info (suppressConsole) {
        suppressConsole = suppressConsole || false;
        return cli.info(this, suppressConsole);
    };

    /**
    * Displays application diagnostic report
    * @function diagnostics
    * @param {boolean} suppressConsole - only returns object instead of writing to console.
    * @augments xipframe
    */
    Xipframe.prototype.diagnostics = function diagnostics (suppressConsole) {
        suppressConsole = suppressConsole || false;
        return cli.diagnostics(this, suppressConsole);
    };

    /**
    * Downloads a file to the specified directory
    * @function info
    * @param {string} source - path to file.
    * @param {string} target - the location to save the file.
    * @param {function} callback - callback function.
    * @param {object} options - sets flags, rename, encoding, reportProgress.
    * @augments xipframe
    */
    Xipframe.prototype.download = function download (source, target, callback, options) {
        utils.diskio.download(source, target, callback, options);
    };

    /**
    * Runs Grunt tasks
    * @function
    * @param {string} task - the pre-defined grunt task to run.
    * @param {function} callback - the callback.
    * @augments xipframe
    */
    Xipframe.prototype.grunt = function grunt (task, callback) {
        var grunt = build.grunt(this);
        if (task) grunt(task, callback);
    };

    /**
    * Runs rollbacks on errors.
    *
    * TODO - create some transaction module for parsed commands.
    *
    * @function
    * @augments xipframe
    */
    Xipframe.prototype.rollback = function () {

        console.log('');
        console.log(' > Rolling back failed transaction.');

        /* make sure we don't have seed data/folders but no installed app */
        if (this.command === this.commands.install) {          
     
            utils.diskio.remove(this.config.paths.app, function (success) {
                if (success) console.log(' > Rollback completed successfully.');
                else console.log(' > Unable to rollback transaction, may require manual fixes.');
                process.exit(1);
            });
        
        }

        /* make sure we exit no matter what */
        process.exit(1);

    }

    Xipframe.prototype.exit = function () {
        process.exit(1);
    };

    return Xipframe;


})();

var instance = null;

function getInstance() {
    
    if (instance === null) {
        instance = new Xipframe();
        instance.constructor = null;
    }
    return instance;    
}

/**
* Xipframe primary module, used to create class instance.
* @module xipframe
*/
exports = module.exports = getInstance();








